@@ -1,14 +1,18 @@
-from __future__ import annotations
-
 import inspect
 import typing
 import uuid
 from dataclasses import dataclass
-from typing import TYPE_CHECKING, Any, Awaitable, Callable, Dict, Final, List, Optional, cast
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Dict,
+    Final,
+    List,
+    Optional,
+    cast,
+)
 
-from robotcode.core.async_tools import threaded
-from robotcode.core.dataclasses import from_dict
-from robotcode.core.logging import LoggingDescriptor
 from robotcode.core.lsp.types import (
     ErrorCodes,
     ExecuteCommandOptions,
@@ -16,16 +20,22 @@ from robotcode.core.lsp.types import (
     LSPAny,
     ServerCapabilities,
 )
+from robotcode.core.utils.dataclasses import from_dict
+from robotcode.core.utils.logging import LoggingDescriptor
 from robotcode.jsonrpc2.protocol import JsonRPCErrorException, rpc_method
-from robotcode.language_server.common.decorators import IsCommand, get_command_name
-from robotcode.language_server.common.has_extend_capabilities import HasExtendCapabilities
-from robotcode.language_server.common.parts.protocol_part import LanguageServerProtocolPart
+from robotcode.language_server.common.decorators import (
+    get_command_id,
+    is_command,
+)
+from robotcode.language_server.common.parts.protocol_part import (
+    LanguageServerProtocolPart,
+)
 
 if TYPE_CHECKING:
     from robotcode.language_server.common.protocol import LanguageServerProtocol
 
 
-_FUNC_TYPE = Callable[..., Awaitable[Optional[LSPAny]]]
+_FUNC_TYPE = Callable[..., Optional[LSPAny]]
 
 
 @dataclass
@@ -34,17 +44,17 @@ class CommandEntry:
     callback: _FUNC_TYPE
 
 
-class CommandsProtocolPart(LanguageServerProtocolPart, HasExtendCapabilities):
+class CommandsProtocolPart(LanguageServerProtocolPart):
     _logger: Final = LoggingDescriptor()
 
     PREFIX: Final = f"{uuid.uuid4()}"
 
-    def __init__(self, parent: LanguageServerProtocol) -> None:
+    def __init__(self, parent: "LanguageServerProtocol") -> None:
         super().__init__(parent)
         self.commands: Dict[str, CommandEntry] = {}
 
     def register(self, callback: _FUNC_TYPE, name: Optional[str] = None) -> str:
-        name = name or get_command_name(callback)
+        name = name or get_command_id(callback)
 
         command = f"{self.PREFIX}.{name}"
 
@@ -60,21 +70,24 @@ class CommandsProtocolPart(LanguageServerProtocolPart, HasExtendCapabilities):
             getattr(instance, k) for k, v in type(instance).__dict__.items() if callable(v) and not k.startswith("_")
         ]
         for method in all_methods:
-            if isinstance(method, IsCommand):
+            if is_command(method):
                 self.register(cast(_FUNC_TYPE, method))
 
     def get_command_name(self, callback: _FUNC_TYPE, name: Optional[str] = None) -> str:
-        name = name or get_command_name(callback)
+        name = name or get_command_id(callback)
 
         return f"{self.PREFIX}.{name}"
 
     def extend_capabilities(self, capabilities: ServerCapabilities) -> None:
         capabilities.execute_command_provider = ExecuteCommandOptions(list(self.commands.keys()))
 
-    @rpc_method(name="workspace/executeCommand", param_type=ExecuteCommandParams)
-    @threaded()
-    async def _workspace_execute_command(
-        self, command: str, arguments: Optional[List[LSPAny]], *args: Any, **kwargs: Any
+    @rpc_method(name="workspace/executeCommand", param_type=ExecuteCommandParams, threaded=True)
+    def _workspace_execute_command(
+        self,
+        command: str,
+        arguments: Optional[List[LSPAny]],
+        *args: Any,
+        **kwargs: Any,
     ) -> Optional[LSPAny]:
         self._logger.debug(lambda: f"execute command {command}")
 
@@ -92,4 +105,4 @@ class CommandsProtocolPart(LanguageServerProtocolPart, HasExtendCapabilities):
                 if i < len(arguments):
                     command_args.append(from_dict(arguments[i], type_hints[i]))
 
-        return await entry.callback(*command_args)
+        return entry.callback(*command_args)
